from collections import OrderedDict
from datetime import datetime
from typing import Mapping, Optional, Union
from urllib.parse import urljoin

import aiohttp
from async_timeout import timeout as _timeout
from dateutil.tz import tzutc
import requests
import simplejson as json

from .auth import generate_signature
from .config import APIConfig, get_config
from .exceptions import SornaAPIError


class Request:

    __slots__ = ['config', 'method', 'path',
                 'data', 'date', 'headers',
                 '_content']

    _allowed_methods = frozenset([
        'GET', 'HEAD', 'POST',
        'PUT', 'PATCH', 'DELETE',
        'OPTIONS'])

    def __init__(self, method: str='GET',
                 path: Optional[str]=None,
                 data: Optional[Mapping]=None,
                 config: Optional[APIConfig]=None):
        self.config = config if config else get_config()
        self.method = method
        if path.startswith('/'):
            path = path[1:]
        self.path = path
        self.data = data if data else OrderedDict()
        self.date = datetime.now(tzutc())
        self.headers = OrderedDict([
            ('Content-Type', 'application/json'),
            ('Date', self.date.isoformat()),
            ('X-Sorna-Version', self.config.version),
        ])
        self._content = None

    @property
    def content(self) -> Union[bytes, bytearray, None]:
        '''
        Retrieves the JSON-encoded body content from request.data object.
        Once called, the result is cached until request.content is set manually.
        After reading this, you need to manually reset it by setting to None
        if you have further content changes of request.data.
        '''
        if self._content is None:
            if not self.data:  # for empty data
                self._content = b''
            else:
                self._content = json.dumps(self.data, ensure_ascii=False).encode()
            self.headers['Content-Length'] = str(len(self._content))
        return self._content

    @content.setter
    def content(self, value: Union[bytes, bytearray, None]):
        '''
        Manually set the raw bytes content of the request body.
        It raises AssertionError if the request already has a value in the `data` field.
        '''
        if value is not None:
            assert not self.data, 'request.data should be empty to set request.content manually.'
            self.headers['Content-Length'] = str(len(value))
        self._content = value

    def sign(self, access_key=None, secret_key=None, hash_type=None):
        '''
        Calculates the signature of the given request and adds the
        Authorization HTTP header.
        It should be called at the very end of request preparation and before
        sending the request to the server.
        '''
        if access_key is None:
            access_key = self.config.access_key
        if secret_key is None:
            secret_key = self.config.secret_key
        if hash_type is None:
            hash_type = self.config.hash_type
        hdrs, _ = generate_signature(
            self.method, self.config.version, self.config.endpoint,
            self.date, self.path, self.content,
            access_key, secret_key, hash_type)
        self.headers.update(hdrs)

    def build_url(self):
        major_ver = self.config.version.split('.', 1)[0]
        path = '/' + self.path if len(self.path) > 0 else ''
        return urljoin(self.config.endpoint, major_ver + path)

    # TODO: attach rate-limit information

    def send(self, sess=None):
        '''
        Sends the request to the server.
        '''
        assert self.method in self._allowed_methods
        if sess is None:
            sess = requests.Session()
        else:
            assert isinstance(sess, requests.Session)
        reqfunc = getattr(sess, self.method.lower())
        resp = reqfunc(self.build_url(),
                       data=self.content,
                       headers=self.headers)
        try:
            return Response(resp.status_code, resp.reason, resp.text,
                            resp.headers['content-type'],
                            resp.headers['content-length'])
        except requests.exceptions.RequestException as e:
            raise SornaAPIError from e

    async def asend(self, sess=None, timeout=10.0):
        '''
        Sends the request to the server.

        This method is a coroutine.
        '''
        assert self.method in self._allowed_methods
        if sess is None:
            sess = aiohttp.ClientSession()
        else:
            assert isinstance(sess, aiohttp.ClientSession)
        with sess:
            reqfunc = getattr(sess, self.method.lower())
            try:
                with _timeout(timeout):
                    resp = await reqfunc(self.build_url(),
                                         data=self.content,
                                         headers=self.headers)
                    async with resp:
                        body = await resp.text()
                        return Response(resp.status, resp.reason, body,
                                        resp.content_type,
                                        len(body))
            except Exception as e:
                raise SornaAPIError from e

    async def connect_websocket(self, sess=None):
        '''
        Creates a WebSocket connection.

        This method is an async-generator, where you can fetch WebSocket
        messages using `async for` clause to the returned value.
        '''
        assert self.method == 'GET'
        if sess is None:
            sess = aiohttp.ClientSession()
        else:
            assert isinstance(sess, aiohttp.ClientSession)
        ws = await sess.ws_connect(self.build_url(), headers=self.headers)
        return ws


class Response:

    __slots__ = ['_status', '_reason',
                 '_content_type', '_content_length',
                 '_body']

    def __init__(self, status, reason, body,
                 content_type='text/plain',
                 content_length=None):
        self._status = status
        self._reason = reason
        self._body = body
        self._content_type = content_type
        self._content_length = content_length

        # TODO: include rate-limiting information from headers

    @property
    def status(self):
        return self._status

    @property
    def reason(self):
        return self._reason

    @property
    def content_type(self):
        return self._content_type

    @property
    def content_length(self):
        return self._content_length

    def text(self):
        return self._body

    def json(self):
        return json.loads(self._body)



import json
import sys
import uuid

import aiohttp
from ..exceptions import SornaAPIError
from ..request import Request


async def create_kernel(kernel_type, client_token=None, max_mem=0, timeout=0, return_id_only=True):
    if client_token is not None:
        assert isinstance(client_token, str)
        assert len(client_token) > 8
    request = Request('POST', '/kernel/create', {
        'lang': kernel_type,
        'clientSessionToken': client_token if client_token else uuid.uuid4().hex,
        'resourceLimits': {
            'maxMem': max_mem,
            'timeout': timeout,
        }
    })
    request.sign()
    resp = await request.asend()
    if resp.status == 201:
        if return_id_only:
            return resp.json()['kernelId']
        return resp.json()
    else:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


async def destroy_kernel(kernel_id):
    request = Request('DELETE', '/kernel/{}'.format(kernel_id))
    request.sign()
    resp = await request.asend()
    if resp.status != 204:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


async def restart_kernel(kernel_id):
    request = Request('PATCH', '/kernel/{}'.format(kernel_id))
    request.sign()
    resp = await request.asend()
    if resp.status != 204:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


async def get_kernel_info(kernel_id):
    request = Request('GET', '/kernel/{}'.format(kernel_id))
    request.sign()
    resp = await request.asend()
    if resp.status == 200:
        return resp.json()
    else:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


async def execute_code(kernel_id, code_id, code):
    request = Request('POST', '/kernel/{}'.format(kernel_id), {
        'codeId': code_id,
        'code': code,
    })
    request.sign()
    resp = await request.asend()
    if resp.status == 200:
        return resp.json()['result']
    else:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


class StreamPty:

    def __init__(self, kernel_id, ws):
        self.kernel_id = kernel_id
        self.ws = ws

    @property
    def closed(self):
        return self.ws.closed

    if sys.version_info < (3, 5, 2):
        async def __aiter__(self):
            return (await self.ws.__aiter__())
    else:
        def __aiter__(self):
            return self.ws.__aiter__()

    async def __anext__(self):
        msg = await self.ws.__anext__()
        return msg

    def send_str(self, raw_str):
        self.ws.send_str(raw_str)

    def resize(self, rows, cols):
        self.ws.send_str(json.dumps({
            'type': 'resize',
            'rows': rows,
            'cols': cols,
        }))

    def restart(self):
        self.ws.send_str(json.dumps({
            'type': 'restart',
        }))

    async def close(self):
        await self.ws.close()


async def stream_pty(kernel_id):
    request = Request('GET', '/stream/kernel/{}/pty'.format(kernel_id))
    request.sign()
    try:
        ws = await request.connect_websocket()
    except aiohttp.errors.HttpProcessingError as e:
        raise SornaAPIError(e.code, e.message)
    return StreamPty(kernel_id, ws)



import uuid

from .exceptions import SornaAPIError
from .request import Request


def create_kernel(kernel_type, client_token=None, max_mem=0, timeout=0, return_id_only=True):
    if client_token is not None:
        assert isinstance(client_token, str)
        assert len(client_token) > 8
    request = Request('POST', '/kernel/create', {
        'lang': kernel_type,
        'clientSessionToken': client_token if client_token else uuid.uuid4().hex,
        'resourceLimits': {
            'maxMem': max_mem,
            'timeout': timeout,
        }
    })
    request.sign()
    resp = request.send()
    if resp.status == 201:
        if return_id_only:
            return resp.json()['kernelId']
        return resp.json()
    else:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


def destroy_kernel(kernel_id):
    request = Request('DELETE', '/kernel/{}'.format(kernel_id))
    request.sign()
    resp = request.send()
    if resp.status != 204:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


def restart_kernel(kernel_id):
    request = Request('PATCH', '/kernel/{}'.format(kernel_id))
    request.sign()
    resp = request.send()
    if resp.status != 204:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


def get_kernel_info(kernel_id):
    request = Request('GET', '/kernel/{}'.format(kernel_id))
    request.sign()
    resp = request.send()
    if resp.status == 200:
        return resp.json()
    else:
        raise SornaAPIError(resp.status, resp.reason, resp.text())


def execute_code(kernel_id, code_id, code):
    request = Request('POST', '/kernel/{}'.format(kernel_id), {
        'codeId': code_id,
        'code': code,
    })
    request.sign()
    resp = request.send()
    if resp.status == 200:
        return resp.json()['result']
    else:
        raise SornaAPIError(resp.status, resp.reason, resp.text())



import os
from typing import Optional

_config = None


class APIConfig:

    DEFAULTS = {
        'endpoint': 'https://api.sorna.io',
        'user_agent': 'Sorna Client Library (Python/v0.1)',
        'version': 'v1.20160915',
    }

    def __init__(self, endpoint: Optional[str]=None,
                 version: str='v1.20160915',
                 user_agent: Optional[str]=None,
                 access_key: Optional[str]=None,
                 secret_key: Optional[str]=None,
                 hash_type: Optional[str]=None):
        self._endpoint = endpoint if endpoint else os.environ.get('SORNA_ENDPOINT', self.DEFAULTS['endpoint'])
        self._version = version if version else self.DEFAULTS['version']
        self._user_agent = user_agent if user_agent else self.DEFAULTS['user_agent']
        self._access_key = access_key if access_key else os.environ['SORNA_ACCESS_KEY']
        self._secret_key = secret_key if secret_key else os.environ['SORNA_SECRET_KEY']
        self._hash_type = hash_type.lower() if hash_type else 'sha256'

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def user_agent(self):
        return self._user_agent

    @property
    def access_key(self):
        return self._access_key

    @property
    def secret_key(self):
        return self._secret_key

    @property
    def version(self):
        return self._version

    @property
    def hash_type(self):
        return self._hash_type


def get_config():
    global _config
    if _config is None:
        _config = APIConfig()
    return _config


def set_config(conf: APIConfig):
    global _config
    _config = conf
